import numpy as np

from fintuna.model.ModelBase import ModelBase, EXECUTION_COSTS


class LongOnly(ModelBase):
    def extract_label(self, data, period):
        returns = data.xs('return', axis=1, level=1)
        next_returns = returns.shift(freq=f'-{period}')
        return next_returns > 0.

    def realized_returns(self, predictions, conf_threshold, returns, period):
        trades = predictions.rank(axis=1, ascending=False, method='first') <= 1
        trades[predictions < conf_threshold] = False

        realized_returns = returns.xs('return', axis=1, level=1)
        realized_returns = realized_returns.mask(~trades, np.nan)
        realized_returns -= EXECUTION_COSTS
        realized_returns = realized_returns.sum(axis=1, min_count=1)

        trades = trades.any(axis=1)
        for trades_p in trades.rolling(period):
            if trades_p.sum() > 1:  # more than one trade within a period is impossible
                trades_p[-1] = False
        realized_returns = realized_returns.mask(~trades, np.nan)
        return realized_returns.resample(period).sum(min_count=1)  # keep nan for empty periods



from abc import ABC, abstractmethod
import pandas as pd

class BaseEnsemble(ABC):
    def __init__(self, models: list, returns_column, period, **kwargs):
        self.models = models
        self.returns_column = returns_column
        self.period = period
        self.asset_ids = set()
        self.features = set()
        for model in models:
            self.asset_ids.update(model.asset_ids)
            self.features.update(model.selected_features)

    # specify which data is passed to sink
    @abstractmethod
    def publish(self, data):
        pass

    @abstractmethod
    def realized_returns(self, data):
        pass

    # average feature importance
    def feature_importances(self):
        feature_importances = []
        for model in self.models:
            feature_importances.append(model.get_feature_importances())
        return pd.DataFrame(feature_importances).mean()

    # average shap values
    def shap_values(self, data):
        shap_values = []
        for model in self.models:
            shap_values.append(model.explain(data))
        shap_values = pd.concat(shap_values, axis=1, keys=range(len(self.models)))
        return shap_values.mean(axis=1, level=1)

    def __len__(self):
        return len(self.models)



import logging
import multiprocessing
import pickle
from typing import Type

import numpy as np
import optuna
import pandas as pd
from optuna.study import StudyDirection

from fintuna.ensemble.IndividualEnsemble import IndividualEnsemble
from fintuna.ensemble.BaseEnsemble import BaseEnsemble
from fintuna.model.ModelBase import ModelBase

n_cores = multiprocessing.cpu_count()
np.random.seed(0)

class FinStudy:
    def __init__(self, Model: Type[ModelBase], data: pd.DataFrame, data_specs: dict = {}, split_specs: dict = {}, name: str = 'fin-study'):

        if 'sampling_freq' not in data_specs:
            if data.index.freqstr:
                self.sampling_freq = data.index.freqstr
            elif 'period' in data_specs:
                self.sampling_freq = data_specs['period']
            else:
                raise ValueError('Could not infer period from data. Please specify a period.')
        if 'period' not in data_specs:
            self.period = self.sampling_freq
        if 'period' in data_specs:
            self.period = data_specs['period']
        if 'sampling_freq' in data_specs:
            self.sampling_freq = data_specs['sampling_freq']

        if data.index.freqstr and pd.Timedelta(data.index.freqstr) != pd.Timedelta(self.sampling_freq):
            logging.warning(
                f'data.index.freq {data.index.freq} divers from sampling_freq {self.sampling_freq}. Using sampling_freq')

        if data.index.inferred_type != 'datetime64':
            raise ValueError('data index not supported. Please use datetime64 dataframe.')

        if not data.index.is_monotonic:
            raise ValueError('data index must be monotonic')
        if 'offset' not in data_specs:
            offsets = data.index - data.index.floor(self.sampling_freq)
            if len(offsets.unique()) == 1:
                self.offset = str(offsets[0]) if offsets[0].total_seconds else None
            else:
                raise ValueError('data has irregular offsets')
        else:
            self.offset = data_specs['offset']

        if 'test_until' in split_specs:
            if type(split_specs['test_until']) is str:
                self.test_until = pd.Timestamp(split_specs['test_until'], tz='UTC')
            if type(split_specs['test_until']) is int:
                self.test_until = data.index[split_specs['test_until']]
            if type(split_specs['test_until']) is float:
                self.test_until = data.index[int(len(data) * split_specs['test_until'])]
        else:
            self.test_until = data.index[int(len(data) * .8)]

        test_len = len(data[:self.test_until])
        if 'train_until' in split_specs:
            if type(split_specs['train_until']) is str:
                self.train_until = pd.Timestamp(split_specs['test_until'], tz='UTC')
            if type(split_specs['train_until']) is int:
                self.train_until = data.index[split_specs['test_until']]
            if type(split_specs['train_until']) is float:
                self.train_until = data.index[int(test_len * split_specs['test_until'])]
        else:
            self.train_until = data.index[int(test_len * .6)]
        if self.train_until > self.test_until:
            raise ValueError('train_until must be smaller than test_until')

        self.returns_column = data_specs['returns_column']
        self.name = name
        self.data = data
        self.Model = Model

    def _create_ensemble(self, trials, data):
        models = []
        for trial in trials:
            model = self.Model(trial, **self.model_params)
            model.train(data, self.period, explainer=True)
            models.append(model)
        return self.ensemble_class(models, self.returns_column, self.period, **self.ensemble_params)

    def get_best_trials(self):
        trials_df = self.study.trials_dataframe()
        ascending = self.study.direction == StudyDirection.MINIMIZE
        best_trial_ids = trials_df.sort_values('value', ascending=ascending)[:self.ensemble_size]['number']
        best_trials = [trial for trial in self.study.get_trials() if trial.number in best_trial_ids]
        return best_trials

    # todo support multiple ensembles
    def explore(self, ensemble_class: Type[BaseEnsemble] = IndividualEnsemble, study_params: dict = None, sampling_params: dict = None, model_params: dict = None, ensemble_size=1, conf_thrs_trials=20, n_trials=100, ensemble_params: dict = None):
        study_params = study_params if study_params else {'direction': 'maximize'}
        if sampling_params:
            if 'n_startup_trials' in sampling_params:
                sampling_params['n_startup_trials'] *= conf_thrs_trials
        else:
            sampling_params = {
                'n_startup_trials': 5 * conf_thrs_trials,
                'seed': 0
                               }
        self.model_params = model_params if model_params else {}
        self.ensemble_params = ensemble_params if ensemble_params else {}
        self.ensemble_class = ensemble_class

        study_params['study_name'] = self.name

        self.ensemble_size = ensemble_size
        self.study = optuna.create_study(**study_params, sampler=optuna.samplers.TPESampler(**sampling_params))

        # train test split
        # exclude one prediction_dur to avoid look ahead bias at feature stacking
        explore_data_train = self.data[:self.train_until - pd.Timedelta(self.period)]
        explore_data_test = self.data[self.train_until:self.test_until]

        for i in range(n_trials):

            tmp_trial = self.study.ask()
            model = self.Model(tmp_trial, **self.model_params)
            model.train(explore_data_train, self.period)

            predictions = model.predict(explore_data_test)
            returns = explore_data_test.loc[:, (model.asset_ids, [self.returns_column])]
            returns = returns - returns.xs(self.returns_column, axis=1, level=1).stack().mean()  # remove trend
            returns = returns.shift(freq=f'-{self.period}').reindex(predictions.index)

            min_prediction, max_prediction = predictions.stack().min(), predictions.stack().max()
            for thrs in np.random.uniform(min_prediction, max_prediction, conf_thrs_trials):
                realized_returns = model.realized_returns(predictions, thrs, returns, self.period)
                # todo make some checks on realized_returns

                performance = model.get_performance(realized_returns)

                # add trial copy to study
                params_thrs = tmp_trial.params.copy()
                params_thrs['conf_thrs'] = thrs
                dist_thrs = tmp_trial.distributions.copy()
                dist_thrs['conf_thrs'] = optuna.distributions.UniformDistribution(min_prediction, max_prediction)
                trial_thrs = optuna.trial.create_trial(params=params_thrs, distributions=dist_thrs, value=performance)
                self.study.add_trial(trial_thrs)
        # select best trials to form an ensemble
        best_trials = self.get_best_trials()

        self.real_data_train = self.data[:self.test_until - pd.Timedelta(self.period)]
        self.real_data_test = self.data[self.test_until:]
        self.ensemble = self._create_ensemble(best_trials, self.real_data_train)

        out_of_sample_realized_returns = self.ensemble.realized_returns(self.real_data_test)
        # calculate exposure
        exposed_dur = (~out_of_sample_realized_returns.isna()).sum() * pd.Timedelta(self.period)
        dur = out_of_sample_realized_returns.index[-1] - out_of_sample_realized_returns.index[0]
        exposure = exposed_dur / dur
        # get exposure adjusted benchmark
        benchmark_realized_returns = self.real_data_test.loc[:, (slice(None), self.returns_column)].mean(axis=1) * exposure

        feature_importances = self.ensemble.feature_importances()
        shap_values = self.ensemble.shap_values(self.real_data_test)
        # todo support asset specific explanations
        return {'feature_importances': feature_importances, 'shap_values': shap_values, 'performance': out_of_sample_realized_returns, 'benchmark': benchmark_realized_returns}


    def finetune(self, n_trials=100):

        for i in range(n_trials):
            trial = self.study.ask()
            model = self.Model(trial, **self.model_params)

            predictions = model.train(self.real_data_train, self.period).predict(self.real_data_test)
            returns_test = self.real_data_test.loc[:, (model.asset_ids, [self.returns_column])]
            returns_test = returns_test.shift(freq=f'-{self.period}').reindex(predictions.index)

            conf_thrs = trial.suggest_uniform('conf_thrs', 0., 1.)
            real_returns = model.realized_returns(predictions, conf_thrs, returns_test, self.period)
            performance = model.get_performance(real_returns)
            self.study.tell(trial, performance)

        best_trials = self.get_best_trials()
        # refit on all data
        self.ensemble = self._create_ensemble(best_trials, self.data)
        self._data_validation = self.data[-10:]  # keep a tiny fraction of data for consistency checks
        self._pub_data_validation = [pd for pd in self.ensemble.publish(self._data_validation[-1:])]  # todo pick periods
        del self.data  # not needed anymore, makes pickeling cheaper
        del self.real_data_train
        del self.real_data_test




from fintuna.model.LongOnly import LongOnly


class Pumps(LongOnly):

    def init_classifier(self):
        clf = super().init_classifier()
        clf.set_params(is_unblance=True)
        return clf

    def extract_label(self, data, period):
        anomaly_magnitude = self.trial.suggest_uniform('anomaly_magnitude', .8, .98)
        returns = data.xs('return', axis=1, level=1)

        thrs = returns.stack().quantile(anomaly_magnitude)
        next_returns = returns.shift(freq=f'-{period}')
        return next_returns > thrs



import pandas as pd

from fintuna.ensemble.BaseEnsemble import BaseEnsemble


class IndividualEnsemble(BaseEnsemble):
    def __init__(self, models: list, returns_column, period, **kwargs):
        super().__init__(models, returns_column, period, **kwargs)
        self.weight = 1 / len(models)  # equally weighted. todo support weighting

    def publish(self, data):
        for i, model in enumerate(self.models):
            yield i, data.index[0], model.predict(data).iloc[0], model.trial.params['conf_thrs'], self.period, self.weight

    def realized_returns(self, data):
        ensemble_realized_returns = []
        for model in self.models:
            predictions = model.predict(data)
            conf_thrs = model.trial.params['conf_thrs']
            returns = data.loc[:, (slice(None), self.returns_column)].shift(freq=f'-{self.period}').reindex(predictions.index)
            realized_returns = model.realized_returns(predictions, conf_thrs, returns, self.period)
            ensemble_realized_returns.append(realized_returns)
        ensemble_realized_returns = pd.DataFrame(ensemble_realized_returns)
        return (ensemble_realized_returns * self.weight).sum()  # equally weighted




from fintuna.finstudy import FinStudy
from fintuna.run import run
import pandas as pd

@pd.api.extensions.register_dataframe_accessor("fin")
class FinAccessor:
    def __init__(self, pandas_obj):
        self._validate(pandas_obj)
        self._obj = pandas_obj

    @staticmethod
    def _validate(obj):
        if not obj.columns.is_unique:
            raise AttributeError("Wrong Dataframe format. Assets must be unique.")
        if obj.columns.nlevels != 2:
            raise AttributeError("Wrong Dataframe format. First level must be assets, second level must be features.")

    @property
    def asset_names(self):
        return self._obj.columns.get_level_values(0).unique()

    @property
    def feature_names(self):
        return self._obj.columns.get_level_values(1).unique()

    def stack_asset_data(self):
        data_stacked = pd.DataFrame()
        for asset_id in self._obj.fin.asset_names:
            data_asset = self._obj[asset_id].copy()
            data_asset['asset_id'] = asset_id
            data_stacked = data_stacked.append(data_asset)
        return data_stacked.sort_index()



import logging
from logging import handlers

import numpy as np
import pandas as pd
from apscheduler.schedulers.blocking import BlockingScheduler

from fintuna import FinStudy

logging.basicConfig(level=logging.DEBUG, filemode='w')  # this must be lowest log level
logging.getLogger('matplotlib.font_manager').disabled = True
log = logging.getLogger('live')
fh_info = handlers.RotatingFileHandler('./logs/ensemble.info')
fh_info.setLevel(logging.INFO)
fh_info.setFormatter(logging.Formatter("%(asctime)s [%(levelname)s] %(message)s"))
log.addHandler(fh_info)


def run(data_func, finstudy: FinStudy, sink):
    # check data_func
    since_val, until_val = finstudy._data_validation.index[0], finstudy._data_validation.index[-1]
    since_val -= pd.Timedelta(finstudy.period)  # data_func is right bound -> first observation at since + period
    data_validation_hat = data_func(since_val, until_val, finstudy._data_validation.fin.asset_names.to_list())
    try:
        pd.testing.assert_frame_equal(data_validation_hat, finstudy._data_validation)
    except AssertionError:
        log.error(f'Data generating process changed: {data_validation_hat} and {finstudy._data_validation} do not match')

    # test finstudy
    for i, pub_data in enumerate(finstudy.ensemble.publish(finstudy._data_validation[-1:])):
        for j, value in enumerate(pub_data):
            try:
                if type(value) is pd.Series:
                    np.testing.assert_allclose(value.values, finstudy._pub_data_validation[i][j].values)
                else:
                    assert value == finstudy._pub_data_validation[i][j]
            except AssertionError:
                log.error('data_func and finstudy are not compatible')

    scheduler = BlockingScheduler(timezone='utc')
    trigger_dur_dt = pd.Timedelta(finstudy.sampling_freq)

    now = pd.Timestamp.utcnow().floor(trigger_dur_dt) + pd.Timedelta(finstudy.offset)
    next_interval = now + trigger_dur_dt

    class LookaheadCheck:
        def __init__(self):
            self.prev_data = pd.DataFrame()

        def __call__(self, *args, **kwargs):
            now = pd.Timestamp.utcnow().floor(finstudy.sampling_freq)
            now += pd.Timedelta(finstudy.offset)
            # get data. also fetch the previous data point for lookahead checks
            since = now - pd.Timedelta(finstudy.period) - pd.Timedelta(finstudy.sampling_freq)
            data = data_func(since, now, finstudy.ensemble.asset_ids)
            # todo do some checks on the data

            cnt_data = data[-1:]
            prev_data_hat = data.iloc[0]  # watch for lookahead bias in the data
            try:
                _prev_data = self.prev_data.loc[prev_data_hat.name]
                identical = np.isclose(prev_data_hat, _prev_data, rtol=.01, equal_nan=True)
                if identical.sum() != len(_prev_data):
                    incorrect = pd.Series(~identical, index=_prev_data.index)
                    log.warning(f'Lookahead bias detected:\n {pd.concat([prev_data_hat.loc[incorrect], _prev_data.loc[incorrect]], axis=1)}')

                self.prev_data = self.prev_data.drop(_prev_data.name)
                self.prev_data = self.prev_data.append(data.iloc[-1])
            except KeyError as e:  # previous data is not available at startup
                log.debug('collecting data for lookahead checks..')
                self.prev_data = self.prev_data.append(data.iloc[-1])
                pass

            for args in finstudy.ensemble.publish(cnt_data):
                sink(*args)

    lookahead_check = LookaheadCheck()
    scheduler.add_job(lookahead_check, 'interval', seconds=int(trigger_dur_dt.total_seconds()),
                  start_date=next_interval.to_pydatetime())
    scheduler.start()



import logging
from abc import ABC, abstractmethod

import lightgbm as lgb
import optuna
import pandas as pd
import sklearn


tune_log = logging.getLogger('tune_log')
live_log = logging.getLogger('live')
pd.options.mode.chained_assignment = None

slippage = .002
fees = .002
EXECUTION_COSTS = fees + slippage
class ModelBase(ABC):

    def __init__(self, trial: optuna.Trial, n_jobs=-1, group_by_asset=None):
        self.trial = trial
        self.clf = None
        self.explainer = None
        self.selected_features = None
        self.n_jobs = n_jobs
        if group_by_asset is not None:
            self.group_by_asset = group_by_asset
        else:
            self.group_by_asset = self.trial.suggest_categorical('group_by_asset', [True, False])

    def select_assets(self, data, period) -> list:
        return data.fin.asset_names

    def init_classifier(self) -> sklearn.base.ClassifierMixin:
        clf = lgb.LGBMClassifier(**{'verbosity': -1,
                                    'objective': 'binary',
                                    'metric': 'binary_logloss',
                                    'learning_rate': .05,
                                    'max_bin': 127,
                                    'random_state': 1
                                    })
        clf.set_params(**{
            'n_estimators': self.trial.suggest_int('n_estimators', 50, 1000),
            'max_depth': self.trial.suggest_int('max_depth', 2, 31),
            'subsample': self.trial.suggest_uniform('subsample', .4, 1.),
            'lambda_l1': self.trial.suggest_loguniform('lambda_l1', .01, 10.),
            'lambda_l2': self.trial.suggest_loguniform('lambda_l2', .01, 10.),
            'num_leaves': self.trial.suggest_int('num_leaves', 2, 63),
            'feature_fraction': self.trial.suggest_uniform('feature_fraction', .5, 1.)
        })
        return clf

    def select_features(self, X_train, y_train, X_estimation, y_estimation, period):
        return X_train.columns

    # build an explainer if selected model doesn't already support shap values
    def create_explainer(self, model, X, y):
        pass

    def shap_values(self, X):
        return self.clf.predict_proba(X, pred_contrib=True)[:, :-1]

    def get_feature_importances(self):
        return pd.Series(self.clf.booster_.feature_importance(importance_type='gain'), index=self.selected_features)

    def _fit(self, X, y):
        if not self.group_by_asset:
            self.selected_features = self.selected_features.drop('asset_id')

        return self.clf.fit(X[self.selected_features], y)

    def _predict_proba(self, X):
        X = X[self.selected_features]
        return pd.DataFrame(self.clf.predict_proba(X)[:, 1], index=X.index)

    def train(self, data, period, explainer=False):
        # self.explainer = None  # if train is called multiple times -> reset explainer
        # assert self.trial['prediction_dur'] == data.attrs['prediction_dur']

        self.asset_ids = self.select_assets(data, period)
        if not len(self.asset_ids):  # todo make some more checks
            raise ValueError('No assets selected')
        data = data[self.asset_ids]

        labels = self.extract_label(data, period)
        if set(labels.columns) != set(data.fin.asset_names):
            raise ValueError('columns of labels do not match assets')

        ## prepare data for ml model
        # merge labels into data
        for asset_id in data.fin.asset_names:
            data[asset_id, 'label'] = labels[asset_id]
        # stack features
        X = data.fin.stack_asset_data()
        no_label = X['label'].isna()
        X, y = X[~no_label].drop('label', axis=1), X[~no_label]['label'].astype(bool)
        # keep categorical features
        categorical_features = data.select_dtypes('category').fin.feature_names.to_list() + ['asset_id']
        X[categorical_features] = X[categorical_features].astype('category')

        split_at = X.index[int(len(X) * .75)]  # using 25% as evaluation data
        # avoid look ahead bias
        X_train, y_train = X[:split_at - pd.Timedelta(period)], y[:split_at - pd.Timedelta(period)]
        X_test, y_test = X[split_at:], y[split_at:]
        self.selected_features = self.select_features(X_train, y_train, X_test, y_test, period)
        selected_features_type = type(self.selected_features)
        if len(self.selected_features) < 1 or len(self.selected_features) > len(data.fin.feature_names) \
                or set(self.selected_features) - set(data.fin.feature_names):
            ValueError('Selected features are invalid')
        if selected_features_type is not pd.Series and selected_features_type is not pd.Index:
            ValueError('Selected features have invalid type')

        self.clf = self.init_classifier()
        self._fit(X[self.selected_features], y)  # refit

        if explainer:
            self.explainer = self.create_explainer(self.clf, X[self.selected_features], y)
        return self  # function chaining

    def predict(self, data):

        data = data[self.asset_ids]
        X = data.fin.stack_asset_data()
        # keep categorical features
        categorical_features = data.select_dtypes('category').fin.feature_names.to_list() + ['asset_id']
        X[categorical_features] = X[categorical_features].astype('category')

        X_asset_ids = X['asset_id']
        predictions = self._predict_proba(X)

        # unstack predictions
        predictions['asset_id'] = X_asset_ids.astype(int)
        predictions = predictions.pivot(columns=['asset_id']).droplevel(0, axis=1)
        return predictions

    def explain(self, data):
        data = data[self.asset_ids]
        X = data.fin.stack_asset_data()
        categorical_features = data.select_dtypes('category').fin.feature_names.to_list() + ['asset_id']
        X[categorical_features] = X[categorical_features].astype('category')
        X = X[self.selected_features]

        # subsampling big data sets
        max_size = 1000
        if len(X) > max_size:
            X = X.sample(max_size, random_state=0).sort_index()
        return pd.DataFrame(self.shap_values(X), columns=self.selected_features, index=X.index)

    @abstractmethod
    def extract_label(self, data, period) -> pd.DataFrame:
        pass

    @abstractmethod
    def realized_returns(self, predictions, conf_threshold, returns, period) -> pd.Series:
        pass

    def get_performance(self, realized_returns) -> float:
        return realized_returns.sum()

